// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#ifndef AISEARCHAGENT_H
#define AISEARCHAGENT_H

#include "BoardEvaluator.h"
#include "TranspositionTable.h"

class Board;
class Move;
class TranspositionTable;
class HistoryTable;

class AISearchAgent
{
/***************************************************************************
* PUBLIC METHODS
**************************************************************************/
public:
  // Construction
  AISearchAgent();
  AISearchAgent(BoardEvaluator& eval);  
  
  // Pick a function which the agent will use to assess the potency of a
  // position.  This may change during the game; for example, a special
  // "mop-up" evaluator may replace the standard when it comes time to drive
  // a decisive advantage home at the end of the game.
  bool AttachEvaluator(BoardEvaluator& eval);
  
  // The basic alpha-beta algorithm, used in one disguise or another by
  // every search agent class
  int AlphaBeta(bool nodeType, Board& theBoard, int depth,
                int alpha, int beta, bool& Continue);    

  // A slight variant of alphabeta which only considers captures and null moves
  // This is necesary because the evaluation function can only be applied to
  // "quiet" positions where the tactical situation (i.e., material balance) is
  // unlikely to change in the near future.
  // Note that, in this version of the code, the quiescence search is not limited
  // by depth; we continue digging for as long as we can find captures.  Some other
  // programs impose a depth limit for time-management purposes.
  int QuiescenceSearch(bool nodeType, Board& theBoard, int alpha, int beta, bool& Continue);   
    
  // Each agent class needs some way of picking a move!
  virtual Move PickBestMove(Board& theBoard, bool& Continue) = 0;

/***************************************************************************
* DATA MEMBERS
**************************************************************************/
protected:
  // A transposition table for this object
  TranspositionTable mTransTable;

  // A handle to the system's history table
  HistoryTable* mHistoryTable;

  // How will we assess position strengths?
  BoardEvaluator Evaluator;
  int FromWhosePerspective;

  // Search node types: MAXNODEs are nodes where the computer player is the
  // one to move; MINNODEs are positions where the opponent is to move.
  static const bool MAXNODE = true;
  static const bool MINNODE = false;

  // Alphabeta search boundaries
  static const int ALPHABETA_MAXVAL = 30000;
  static const int ALPHABETA_MINVAL = -30000;
  static const int ALPHABETA_ILLEGAL = -31000;

  // An approximate upper bound on the total value of all positional
  // terms in the evaluation function
  static const int EVAL_THRESHOLD = 200;

  // A score below which we give up: if Alphabeta ever returns a value lower
  // than this threshold, then all is lost and we might as well resign.  Here,
  // the value is equivalent to "mated by the opponent in 3 moves or less".
  static const int ALPHABETA_GIVEUP = -29995;

public:
  // ID's for concrete subclasses; AISearchAgent works as a factory for its
  // concrete subclasses
  static const int AISEARCH_ALPHABETA = 0;
  static const int AISEARCH_MTDF = 1;  

  // Statistics
  int NumRegularNodes;
  int NumQuiescenceNodes;
  int NumRegularTTHits;
  int NumQuiescenceTTHits;
  int NumRegularCutoffs;
  int NumQuiescenceCutoffs;

  // A move counter, so that the agent knows when it can delete old stuff from
  // its transposition table
  int MoveCounter;
};

#endif
