// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#ifndef MOVELISTGENERATOR_H
#define MOVELISTGENERATOR_H

class Move;
class Board;

class MoveListGenerator
{
public:
  // Construction
  MoveListGenerator();
  ~MoveListGenerator();

  // Prepare an iterator for scanning through the list of moves
  void ResetIterator(void);
  
  // Accessors
  std::vector<Move> GetMoveList(void);
  size_t Size(void);
  
  // Look for a specific move in the list; if it is there, return true
  // This is used by the PlayerHuman object, to verify whether a move entered
  // by the player is actually valid
  bool Find(Move& mov);
  
    // look for a move from "source" to "dest" in the list
  Move FindMoveForSquares(int source, int dest);
  
  // Find the next move in the list
  std::vector<Move>::iterator Next(void);
  
  // Get the end of the list
  std::vector<Move>::iterator End(void) { return Moves.end(); }
  
  // Look at the board received as a parameter, and build a list of legal
  // moves which can be derived from it.  If there are no legal moves, or if
  // one of the moves is a king capture (which means that the opponent's
  // previous move left the king in check, which is illegal), return false.
  bool ComputeLegalMoves(Board& theBoard, bool& Continue); 
  
  // Find only the moves which are relevant to quiescence search; i.e., captures
  bool ComputeQuiescenceMoves(Board& theBoard, bool& Continue);  

  // Dump the move list to standard output, for debugging purposes
  void Print(void);

  static void Initialize(void);
  
private:
   bool ComputeWhiteQueenMoves(Board& theBoard, bool& Continue);
   bool ComputeWhiteKingMoves(Board& theBoard, bool& Continue);      
   bool ComputeWhiteRookMoves(Board& theBoard, int pieceType, bool& Continue);   
   bool ComputeWhiteBishopMoves(Board& theBoard, int pieceType, bool& Continue);  
   bool ComputeWhiteKnightMoves(Board& theBoard, bool& Continue);
   bool ComputeWhitePawnMoves(Board& theBoard, bool& Continue);

   bool ComputeBlackQueenMoves(Board& theBoard, bool& Continue);
   bool ComputeBlackKingMoves(Board& theBoard, bool& Continue);      
   bool ComputeBlackRookMoves(Board& theBoard, int pieceType, bool& Continue);   
   bool ComputeBlackBishopMoves(Board& theBoard, int pieceType, bool& Continue);  
   bool ComputeBlackKnightMoves(Board& theBoard, bool& Continue);
   bool ComputeBlackPawnMoves(Board& theBoard, bool& Continue);         
  
  // Pre-processed data structures containing all possible moves from all
  // possible squares, by piece type
  static int KnightMoves[64][8];
  static int KnightLengths[64];

  static int KingMoves[64][8];
  static int KingLengths[64];  

  static int BishopMoves[64][4][7];
  static int BishopOuterLengths[64];
  static int BishopInnerLengths[64][4];

  static int RookMoves[64][4][7];      
  static int RookOuterLengths[64];
  static int RookInnerLengths[64][4];    

  static bool IsInitialized;
  
public:
  std::vector<Move> Moves;  
  std::vector<Move>::iterator MovesIt;  

};

#endif
