// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "AISearchAgentAlphabeta.h"
#include "Move.h"
#include "Board.h"
#include "HistoryTable.h"
#include "MoveListGenerator.h"


// Implementation of the abstract method defined in the superclass
Move AISearchAgentAlphabeta::PickBestMove(Board& theBoard, bool& Continue)
{
  // Store the identity of the moving side, so that we can tell Evaluator
  // from whose perspective we need to evaluate positions
  FromWhosePerspective = theBoard.GetCurrentPlayer();
  MoveCounter++;

  // Should we erase the history table?
  if ((rand()%4)==2)
    (*mHistoryTable).Forget();

  NumRegularNodes = 0; NumQuiescenceNodes = 0;
  NumRegularTTHits = 0; NumQuiescenceTTHits = 0;

  // Find the moves
  Move theMove;
  MoveListGenerator movegen;
  movegen.ComputeLegalMoves(theBoard, Continue);
  (*mHistoryTable).SortMoveList(movegen, theBoard.GetCurrentPlayer());

  // The following code blocks look a lot like the max node case from
  // AISearchAgent.Alphabeta, with an added twist: we need to store the
  // actual best move, and not only pass around its minimax value
  int bestSoFar = ALPHABETA_MINVAL;
  Board newBoard;
  Move mov;
  int currentAlpha = ALPHABETA_MINVAL;

  // Loop on all pseudo-legal moves
  std::vector<Move>::iterator testMove;
  while ((testMove = movegen.Next()) != movegen.End())    
  {
    newBoard.Clone(theBoard);
    newBoard.ApplyMove(*testMove);
    int movScore = AlphaBeta(MINNODE, newBoard, 5, currentAlpha, ALPHABETA_MAXVAL, Continue);
    if (movScore == ALPHABETA_ILLEGAL)
      continue;

    currentAlpha = max(currentAlpha, movScore);

    if (movScore > bestSoFar)
    {
      theMove.Copy(*testMove);
      bestSoFar = movScore;
      theMove.MoveEvaluation = movScore;
    }
  }

  // And now, if the best we can do is ALPHABETA_GIVEUP or worse, then it is
  // time to resign...  Unless the opponent was kind wnough to put us in
  // stalemate!
  if (bestSoFar <= ALPHABETA_GIVEUP)
  {
    // Check for a stalemate
    // Stalemate occurs if the player's king is NOT in check, but all of his
    // moves are illegal.
    // First, verify whether we are in check
    newBoard.Clone(theBoard);
    MoveListGenerator secondary;
    newBoard.SwitchSides();
    if (secondary.ComputeLegalMoves(newBoard, Continue))
    {
      // Then, we are not in check and may continue our efforts.
      // We must now examine all possible moves; if at least one resuls in
      // a legal position, there is no stalemate and we must assume that
      // we are doomed
      (*mHistoryTable).SortMoveList(movegen, newBoard.GetCurrentPlayer());
      movegen.ResetIterator();
      // If we can scan all moves without finding one which results
      // in a legal position, we have a stalemate
      theMove.MoveType = Move::MOVE_STALEMATE;
      theMove.MovingPiece = Board::KING + theBoard.GetCurrentPlayer();
      // Look at the moves
      while ((testMove = movegen.Next()) != movegen.End())    
      {
        newBoard.Clone(theBoard);
        newBoard.ApplyMove(*testMove);
        if (secondary.ComputeLegalMoves(newBoard, Continue))
        {
          theMove.MoveType = Move::MOVE_RESIGN;
        }
      }
    } else {
      // We're in check and our best hope is GIVEUP or worse, so either we are
      // already checkmated or will be soon, without hope of escape
      theMove.MovingPiece = Board::KING + theBoard.GetCurrentPlayer();
      theMove.MoveType = Move::MOVE_RESIGN;
    }
  }
  
  return theMove;
}