// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "Board.h"
#include "Move.h"
#include "Hand.h"
#include "HandState.h"
#include "StateManager.h"
#include "GameOverState.h"
#include "MoveListGenerator.h"

dong Board::SquareBits[ALL_SQUARES];

int Board::HashKeyComponents[ALL_SQUARES][ALL_PIECES];
int Board::HashLockComponents[ALL_SQUARES][ALL_PIECES];

String Board::PieceStrings[ALL_PIECES + 1];
int Board::PieceValues[ALL_PIECES];

dong Board::EXTRAKINGS_WHITE_KINGSIDE = 0;
dong Board::EXTRAKINGS_WHITE_QUEENSIDE = 0;
dong Board::EXTRAKINGS_BLACK_KINGSIDE = 0;
dong Board::EXTRAKINGS_BLACK_QUEENSIDE = 0;
dong Board::EMPTYSQUARES_WHITE_KINGSIDE = 0;
dong Board::EMPTYSQUARES_WHITE_QUEENSIDE = 0;
dong Board::EMPTYSQUARES_BLACK_KINGSIDE = 0;
dong Board::EMPTYSQUARES_BLACK_QUEENSIDE = 0;

bool Board::IsInitialized = false;

Board::Board()
{         
  if (!IsInitialized)
    Initialize();   

  StartingBoard(); 
}

Board::Board(Board* original)
{
  Clone(*original);
}

Board::~Board() {}

// Restore the board to a game-start position
bool Board::StartingBoard()
{
  // Put the pieces on the board
  EmptyBoard();  
  
  // load the bit boards
  AddPiece(0, BLACK_ROOK);
  AddPiece(1, BLACK_KNIGHT);
  AddPiece(2, BLACK_BISHOP);
  AddPiece(3, BLACK_QUEEN);
  AddPiece(4, BLACK_KING);
  AddPiece(5, BLACK_BISHOP);
  AddPiece(6, BLACK_KNIGHT);
  AddPiece(7, BLACK_ROOK);

  for (int i = 8; i < 16; i++) {
    AddPiece(i, BLACK_PAWN);
  }

  for (int i = 48; i < 56; i++) {
    AddPiece(i, WHITE_PAWN);
  }

  AddPiece(56, WHITE_ROOK);
  AddPiece(57, WHITE_KNIGHT);
  AddPiece(58, WHITE_BISHOP);
  AddPiece(59, WHITE_QUEEN);
  AddPiece(60, WHITE_KING);
  AddPiece(61, WHITE_BISHOP);
  AddPiece(62, WHITE_KNIGHT);
  AddPiece(63, WHITE_ROOK);
  
  // And allow all castling moves
  for (int i = 0; i < 4; i++) {
    CastlingStatus[i] = true;
  }

  HasCastled[0] = false;
  HasCastled[1] = false;

  ClearEnPassantPawn();

  // And ask White to play the first move
  SetCurrentPlayer(Player::SIDE_WHITE);   
  
  return true;
}

// Remove every piece from the board
bool Board::EmptyBoard()
{
  for(int i = 0; i < ALL_BITBOARDS; i++ ) {
    BitBoards[i] = 0;
  }
  ExtraKings[0] = 0;
  ExtraKings[1] = 0;
  EnPassantPawn = 0;
  MaterialValue[0] = 0;
  MaterialValue[1] = 0;
  NumPawns[0] = 0;
  NumPawns[1] = 0;
  return true;
}

// Place a specific piece on a specific board square
bool Board::AddPiece(int whichSquare, int whichPiece)
{
  // Add the piece itself
  BitBoards[whichPiece] |= SquareBits[whichSquare];

  // And note the new piece position in the bitboard containing all
  // pieces of its color.  Here, we take advantage of the fact that
  // all pieces of a given color are represented by numbers of the same
  // parity
  BitBoards[ALL_PIECES + (whichPiece % 2)] |= SquareBits[whichSquare];

  // And adjust material balance accordingly
  MaterialValue[whichPiece % 2] += PieceValues[whichPiece];
  if (whichPiece == WHITE_PAWN)
    NumPawns[Player::SIDE_WHITE]++;
  else if (whichPiece == BLACK_PAWN )
    NumPawns[ Player::SIDE_BLACK ]++;
  
  return true;  
}

// Eliminate a specific piece from a specific square on the board
// Note that you MUST know that the piece is there before calling this,
// or the results will not be what you expect!
bool Board::RemovePiece(int whichSquare, int whichPiece)
{
  // Remove the piece itself
  BitBoards[whichPiece] ^= SquareBits[whichSquare];
  BitBoards[ALL_PIECES + (whichPiece%2)] ^= SquareBits[whichSquare];

  // And adjust material balance accordingly
  MaterialValue[whichPiece%2] -= PieceValues[whichPiece];
  if (whichPiece == WHITE_PAWN)
    NumPawns[Player::SIDE_WHITE]--;
  else if (whichPiece == BLACK_PAWN)
    NumPawns[Player::SIDE_BLACK]--;
  
  return true;
}

// static member initialization
void Board::Initialize(void)
{     
  // Build the SquareBits constants  
  for(int i = 0; i < ALL_SQUARES; i++)
  {
    // Note: the dong(1) specifies that the 1 we are shifting is a long dong
    // Otherwise, by default, it would be a 4-byte int and be unable to
    // shift the 1 to bits 32 to 63   
    SquareBits[i] = ( dong(1) << i );
  }
  
  // Build the extrakings constants
  EXTRAKINGS_WHITE_KINGSIDE = SquareBits[60] | SquareBits[61];
  EXTRAKINGS_WHITE_QUEENSIDE = SquareBits[60] | SquareBits[59];
  EXTRAKINGS_BLACK_KINGSIDE = SquareBits[4] | SquareBits[5];
  EXTRAKINGS_BLACK_QUEENSIDE = SquareBits[4] | SquareBits[3];
  EMPTYSQUARES_WHITE_KINGSIDE = SquareBits[61] | SquareBits[62];
  EMPTYSQUARES_WHITE_QUEENSIDE = SquareBits[59] | SquareBits[58] | SquareBits[57];
  EMPTYSQUARES_BLACK_KINGSIDE = SquareBits[5] | SquareBits[6];
  EMPTYSQUARES_BLACK_QUEENSIDE = SquareBits[3] | SquareBits[2] | SquareBits[1];
  
  // Build the hashing database   
  for(int i = 0; i < ALL_PIECES; i++)
  {
    for(int j = 0; j < ALL_SQUARES; j++)
    {
      HashKeyComponents[i][j] = rand();
      HashLockComponents[i][j] = rand();
    }
  }

  // Tokens representing the various concepts in the game, for printint
  // and file i/o purposes
  // PieceStrings contains an extra String representing empty squares
  PieceStrings[WHITE_PAWN] = "WP";
  PieceStrings[WHITE_ROOK] = "WR";
  PieceStrings[WHITE_KNIGHT] = "WN";
  PieceStrings[WHITE_BISHOP] = "WB";
  PieceStrings[WHITE_QUEEN] = "WQ";
  PieceStrings[WHITE_KING] = "WK";
  PieceStrings[BLACK_PAWN] = "BP";
  PieceStrings[BLACK_ROOK] = "BR";
  PieceStrings[BLACK_KNIGHT] = "BN";
  PieceStrings[BLACK_BISHOP] = "BB";
  PieceStrings[BLACK_QUEEN] = "BQ";
  PieceStrings[BLACK_KING] = "BK";
  PieceStrings[ALL_PIECES] = "  ";
  
  // Numerical evaluation of piece material values
  PieceValues[WHITE_PAWN] = 100;
  PieceValues[BLACK_PAWN] = 100;
  PieceValues[WHITE_KNIGHT] = 300;
  PieceValues[BLACK_KNIGHT] = 300;
  PieceValues[WHITE_BISHOP] = 350;
  PieceValues[BLACK_BISHOP] = 350;
  PieceValues[WHITE_ROOK] = 500;
  PieceValues[BLACK_ROOK] = 500;
  PieceValues[BLACK_QUEEN] = 900;
  PieceValues[WHITE_QUEEN] = 900;
  PieceValues[WHITE_KING] = 2000;
  PieceValues[BLACK_KING] = 2000;   

  IsInitialized = true;
}

// Indicates that there is no en passant square at all.  Technically, this
// job could have been handled by SetEnPassantPawn( long ) with a null
// parameter, but I have chosen to add a method to avoid problems if I ever
// forgot to specify 0L: using 0 would call the first form of the Set method
// and indicate an en passant pawn in a corner of the board, with possibly
// disastrous consequences!
bool Board::ClearEnPassantPawn(void)
{
  EnPassantPawn = 0;
  return true;
}

// Whose turn is it?
bool Board::SetCurrentPlayer(int which)
{
  CurrentPlayer = which;
  return true;
}

// Change the identity of the player to move
int Board::SwitchSides(void)
{
  if (CurrentPlayer == Player::SIDE_WHITE)
    SetCurrentPlayer(Player::SIDE_BLACK);
  else
    SetCurrentPlayer(Player::SIDE_WHITE);
  return CurrentPlayer;
}

bool Board::ApplyMove(const Move& theMove, const bool ConfirmMove)
{  
  // If the move includes a pawn promotion, an extra step will be required
  // at the end
  bool isPromotion = ( theMove.MoveType >= Move::MOVE_PROMOTION_KNIGHT );
  int moveWithoutPromotion = ( theMove.MoveType & Move::NO_PROMOTION_MASK );
  int side = theMove.MovingPiece % 2;
  int theRook = 0;
  bool capture = false;

  // For now, ignore pawn promotions
  switch (moveWithoutPromotion)
  {
    case (Move::MOVE_NORMAL):
      // The simple case
      if (ConfirmMove) {
        Hand::GetHand(side)->SetAction(HandState::Move, theMove.SourceSquare, theMove.DestinationSquare);
      }
      RemovePiece( theMove.SourceSquare, theMove.MovingPiece );
      AddPiece( theMove.DestinationSquare, theMove.MovingPiece );      
      break;
    case (Move::MOVE_CAPTURE_ORDINARY):
      // Don't forget to remove the captured piece!
      if (ConfirmMove) {
        Hand::GetHand(side)->SetAction(HandState::Capture, theMove.SourceSquare, theMove.DestinationSquare);        
        capture = true;
      }
      RemovePiece( theMove.SourceSquare, theMove.MovingPiece );
      RemovePiece( theMove.DestinationSquare, theMove.CapturedPiece );
      AddPiece( theMove.DestinationSquare, theMove.MovingPiece );
      break;
    case (Move::MOVE_CAPTURE_EN_PASSANT):      
      // move the hands
      if (ConfirmMove) {
        if ( ( theMove.MovingPiece % 2 ) == Player::SIDE_WHITE )
          Hand::GetHand(side)->SetAction(HandState::EnPassant, theMove.SourceSquare, theMove.DestinationSquare);        
        else
          Hand::GetHand(side)->SetAction(HandState::EnPassant, theMove.SourceSquare, theMove.DestinationSquare);        
      }
      // Here, we can use our knowledge of the board to make a small
      // optimization, since the pawn to be captured is always
      // "behind" the moving pawn's destination square, we can compute its
      // position on the fly
      RemovePiece( theMove.SourceSquare, theMove.MovingPiece );
      AddPiece( theMove.DestinationSquare, theMove.MovingPiece );
      if ( ( theMove.MovingPiece % 2 ) == Player::SIDE_WHITE )
        RemovePiece( theMove.DestinationSquare + 8, theMove.CapturedPiece );
      else
        RemovePiece( theMove.DestinationSquare - 8, theMove.CapturedPiece );
      break;
    case (Move::MOVE_CASTLING_QUEENSIDE):
      // move the hands
      if (ConfirmMove) {        
        Hand::GetHand(side)->SetAction(HandState::Castle, theMove.SourceSquare, theMove.DestinationSquare, theMove.SourceSquare - 4, theMove.SourceSquare - 1);                
      }
      // Again, we can compute the rook's source and destination squares
      // because of our knowledge of the board's structure
      RemovePiece( theMove.SourceSquare, theMove.MovingPiece );
      AddPiece( theMove.DestinationSquare, theMove.MovingPiece );
      theRook = ROOK + ( theMove.MovingPiece % 2 );
      RemovePiece( theMove.SourceSquare - 4, theRook );
      AddPiece( theMove.SourceSquare - 1, theRook );
      // We must now mark some squares as containing "phantom kings" so that
      // the castling can be cancelled by the next opponent's move, if he
      // can move to one of them
      if ( side == Player::SIDE_WHITE ) {
        SetExtraKings( side, EXTRAKINGS_WHITE_QUEENSIDE );
      } else {
        SetExtraKings( side, EXTRAKINGS_BLACK_QUEENSIDE );
      }
      HasCastled[side] = true;
      break;
    case (Move::MOVE_CASTLING_KINGSIDE):
      if (ConfirmMove) {        
        Hand::GetHand(side)->SetAction(HandState::Castle, theMove.SourceSquare, theMove.DestinationSquare, theMove.SourceSquare + 3, theMove.SourceSquare + 1);                
      }
      // Again, we can compute the rook's source and destination squares
      // because of our knowledge of the board's structure
      RemovePiece( theMove.SourceSquare, theMove.MovingPiece );
      AddPiece( theMove.DestinationSquare, theMove.MovingPiece );
      theRook = ROOK + ( theMove.MovingPiece % 2 );
      RemovePiece( theMove.SourceSquare + 3, theRook );
      AddPiece( theMove.SourceSquare + 1, theRook );
      // We must now mark some squares as containing "phantom kings" so that
      // the castling can be cancelled by the next opponent's move, if he
      // can move to one of them
      if ( side == Player::SIDE_WHITE ) {
        SetExtraKings(side, EXTRAKINGS_WHITE_KINGSIDE );
      } else {
        SetExtraKings(side, EXTRAKINGS_BLACK_KINGSIDE );
      }
      HasCastled[side] = true;
      break;
    case (Move::MOVE_RESIGN):
      if (ConfirmMove) {
        if (side == Player::SIDE_WHITE) {
          GameOverState::getSingletonPtr()->DisplayResult("White Resigns", "Black Wins!");            
        } else {
          GameOverState::getSingletonPtr()->DisplayResult("Black Resigns", "White Wins!");            
        }
        StateManager::getSingleton().pushState(GameOverState::getSingletonPtr());
      }
      break;
    case (Move::MOVE_STALEMATE):
      if (ConfirmMove) {
        GameOverState::getSingletonPtr()->DisplayResult("Stalemate", "Match Drawn");          
        StateManager::getSingleton().pushState(GameOverState::getSingletonPtr());
      }
      break;
  }

  // And now, apply the promotion
  if (isPromotion)
  {
    int promotionType = (theMove.MoveType & Move::PROMOTION_MASK);
    int color = (theMove.MovingPiece % 2);    
    switch (promotionType)
    {
      case (Move::MOVE_PROMOTION_KNIGHT):                 
        if (ConfirmMove) {
          // do the hand movement
          if (capture)
            Hand::GetHand(side)->SetAction(HandState::Capture, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Knight");        
          else
            Hand::GetHand(side)->SetAction(HandState::Promotion, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Knight");        
        }
        RemovePiece( theMove.DestinationSquare, theMove.MovingPiece );
        AddPiece( theMove.DestinationSquare, KNIGHT + color );
        break;
      case (Move::MOVE_PROMOTION_BISHOP):
        if (ConfirmMove) {
          // do the hand movement
          if (capture)
            Hand::GetHand(side)->SetAction(HandState::Capture, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Bishop");        
          else
            Hand::GetHand(side)->SetAction(HandState::Promotion, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Bishop");        
        }
        RemovePiece( theMove.DestinationSquare, theMove.MovingPiece );
        AddPiece( theMove.DestinationSquare, BISHOP + color );
        break;
      case (Move::MOVE_PROMOTION_ROOK):
        if (ConfirmMove) {
          // do the hand movement
          if (capture)
            Hand::GetHand(side)->SetAction(HandState::Capture, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Rook");        
          else
            Hand::GetHand(side)->SetAction(HandState::Promotion, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Rook");        
        }
        RemovePiece( theMove.DestinationSquare, theMove.MovingPiece );
        AddPiece( theMove.DestinationSquare, ROOK + color );
        break;
      case (Move::MOVE_PROMOTION_QUEEN):
        if (ConfirmMove) {
          // do the hand movement
          if (capture)
            Hand::GetHand(side)->SetAction(HandState::Capture, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Queen");        
          else
            Hand::GetHand(side)->SetAction(HandState::Promotion, theMove.SourceSquare, theMove.DestinationSquare, -1, -1, "Queen");        
        }
        RemovePiece( theMove.DestinationSquare, theMove.MovingPiece );
        AddPiece( theMove.DestinationSquare, QUEEN + color );
        break;
    }    
  }

  // If this was a 2-step pawn move, we now have a valid en passant
  // capture possibility.  Otherwise, no.
  if ( ( theMove.MovingPiece == Board::WHITE_PAWN ) &&
        ( theMove.SourceSquare - theMove.DestinationSquare == 16 ) )
    SetEnPassantPawn( int(theMove.DestinationSquare + 8) );
  else if ( ( theMove.MovingPiece == Board::BLACK_PAWN ) &&
            ( theMove.DestinationSquare - theMove.SourceSquare == 16 ) )
    SetEnPassantPawn( int(theMove.SourceSquare + 8) );
  else
    ClearEnPassantPawn();

  // And now, maintain castling status
  // If a king moves, castling becomes impossible for that side, for the
  // rest of the game
  switch( theMove.MovingPiece )
  {
    case WHITE_KING:
      SetCastlingStatus( CASTLE_KINGSIDE + Player::SIDE_WHITE, false );
      SetCastlingStatus( CASTLE_QUEENSIDE + Player::SIDE_WHITE, false );
      break;
    case BLACK_KING:
      SetCastlingStatus( CASTLE_KINGSIDE + Player::SIDE_BLACK, false );
      SetCastlingStatus( CASTLE_QUEENSIDE + Player::SIDE_BLACK, false );
      break;
    default:
      break;
  }

  // Or, if ANYTHING moves from a corner, castling becomes impossible on
  // that side (either because it's the rook that is moving, or because
  // it has been captured by whatever moves, or because it is already gone)
  switch( theMove.SourceSquare )
  {
    case 0:
      SetCastlingStatus( CASTLE_QUEENSIDE + Player::SIDE_BLACK, false );
      break;
    case 7:
      SetCastlingStatus( CASTLE_KINGSIDE + Player::SIDE_BLACK, false );
      break;
    case 56:
      SetCastlingStatus( CASTLE_QUEENSIDE + Player::SIDE_WHITE, false );
      break;
    case 63:
      SetCastlingStatus( CASTLE_KINGSIDE + Player::SIDE_WHITE, false );
      break;
    default:
      break;
  }

  // All that remains to do is switch sides
  SetCurrentPlayer((GetCurrentPlayer()+1)%2);
  return true;
}

// Make a deep copy of a Board object; assumes that memory has already
// been allocated for the new object, which is always true since we
// "allocate" Boards from a permanent array
bool Board::Clone(const Board& target )
{
  EnPassantPawn = target.EnPassantPawn;
  for (int i = 0; i < 4; i++) {
    CastlingStatus[i] = target.CastlingStatus[i];
  }
  for (int i = 0; i < ALL_BITBOARDS; i++) {
    BitBoards[i] = target.BitBoards[i];
  }
  MaterialValue[0] = target.MaterialValue[0];
  MaterialValue[1] = target.MaterialValue[1];
  NumPawns[0] = target.NumPawns[0];
  NumPawns[1] = target.NumPawns[1];
  ExtraKings[0] = target.ExtraKings[0];
  ExtraKings[1] = target.ExtraKings[1];
  HasCastled[0] = target.HasCastled[0];
  HasCastled[1] = target.HasCastled[1];
  CurrentPlayer = target.CurrentPlayer;
  return true;
}

// Compute a 32-bit integer to represent the board, according to Zobrist[70]
int Board::HashKey(void)
{
  int hash = 0;
  // Look at all pieces, one at a time
  for(int currPiece = 0; currPiece < ALL_PIECES; currPiece++)
  {
    dong tmp = BitBoards[currPiece];
    // Search for all pieces on all squares.  We could optimize here: not
    // looking for pawns on the back row (or the eight row), getting out
    // of the "currSqaure" loop once we found one king of one color, etc.
    // But for simplicity's sake, we'll keep things generic.
    for (int currSquare = 0; currSquare < ALL_SQUARES; currSquare++)
    {
      // Zobrist's method: generate a bunch of random bitfields, each
      // representing a certain "piece X is on square Y" predicate; XOR
      // the bitfields associated with predicates which are true.
      // Therefore, if we find a piece (in tmp) in a certain square,
      // we accumulate the related HashKeyComponent.
      if ((tmp & SquareBits[currSquare]) != 0)
        hash ^= HashKeyComponents[currPiece][currSquare];
    }
  }
  return hash;
}

// Compute a second 32-bit hash key, using an entirely different set
// piece/square components.
// This is required to be able to detect hashing collisions without
// storing an entire jcBoard in each slot of the TranspositionTable,
// which would gobble up inordinate amounts of memory
int Board::HashLock(void)
{
  int hash = 0;
  for (int currPiece = 0; currPiece < ALL_PIECES; currPiece++)
  {
    dong tmp = BitBoards[currPiece];
    for (int currSquare = 0; currSquare < ALL_SQUARES; currSquare++)
    {
      if ((tmp & SquareBits[currSquare]) != 0)
        hash ^= HashLockComponents[currPiece][currSquare];
    }
  }
  return hash;
}

void Board::SetExtraKings(int side, dong val)
{
  // Mark a few squares as containing "phantom kings" to detect illegal
  // castling
  ExtraKings[side] = val;
  BitBoards[KING + side] |= ExtraKings[side];
  BitBoards[ALL_PIECES + side] |= ExtraKings[side];
}

void Board::ClearExtraKings(int side)
{
  BitBoards[KING + side] ^= ExtraKings[side];
  BitBoards[ALL_PIECES + side] ^= ExtraKings[side];
  // Note: one of the Extra Kings is superimposed on the rook involved in
  // the castling, so the next step is required to prevent ALL_PIECES from
  // forgetting about the rook at the same time as the phantom king
  BitBoards[ALL_PIECES + side] |= BitBoards[ROOK + side];
  ExtraKings[side] = 0;
}

// Change one of the "castling status" flags
// parameter whichFlag should be a sum of a side marker and a castling
// move identifier, for example, Player.SIDE_WHITE + CASTLE_QUEENSIDE
bool Board::SetCastlingStatus( int whichFlag, bool newValue )
{
  CastlingStatus[ whichFlag ] = newValue;
  return true;
}

// If a pawn move has just made en passant capture possible, mark it as
// such in a bitboard (containing the en passant square only)
bool Board::SetEnPassantPawn( int square )
{
  ClearEnPassantPawn();
  EnPassantPawn |= SquareBits[ square ];
  return true;
}

bool Board::SetEnPassantPawn( dong bitboard )
{
  EnPassantPawn = bitboard;
  return true;
}

// Look for the piece located on a specific square
int Board::FindBlackPiece( int square )
{
  // Note: we look for kings first for two reasons: because it helps
  // detect check, and because there may be a phantom king (marking an
  // illegal castling move) and a rook on the same square!
  if ( ( BitBoards[ BLACK_KING ] & SquareBits[ square ] ) != 0 )
    return BLACK_KING;
  if ( ( BitBoards[ BLACK_QUEEN ] & SquareBits[ square ] ) != 0 )
    return BLACK_QUEEN;
  if ( ( BitBoards[ BLACK_ROOK ] & SquareBits[ square ] ) != 0 )
    return BLACK_ROOK;
  if ( ( BitBoards[ BLACK_KNIGHT ] & SquareBits[ square ] ) != 0 )
    return BLACK_KNIGHT;
  if ( ( BitBoards[ BLACK_BISHOP ] & SquareBits[ square ] ) != 0 )
    return BLACK_BISHOP;
  if ( ( BitBoards[ BLACK_PAWN ] & SquareBits[ square ] ) != 0 )
    return BLACK_PAWN;
  return EMPTY_SQUARE;
}
int Board::FindWhitePiece( int square )
{
  if ( ( BitBoards[ WHITE_KING ] & SquareBits[ square ] ) != 0 )
    return WHITE_KING;
  if ( ( BitBoards[ WHITE_QUEEN ] & SquareBits[ square ] ) != 0 )
    return WHITE_QUEEN;
  if ( ( BitBoards[ WHITE_ROOK ] & SquareBits[ square ] ) != 0 )
    return WHITE_ROOK;
  if ( ( BitBoards[ WHITE_KNIGHT ] & SquareBits[ square ] ) != 0 )
    return WHITE_KNIGHT;
  if ( ( BitBoards[ WHITE_BISHOP ] & SquareBits[ square ] ) != 0 )
    return WHITE_BISHOP;
  if ( ( BitBoards[ WHITE_PAWN ] & SquareBits[ square ] ) != 0 )
    return WHITE_PAWN;
  return EMPTY_SQUARE;
}

// Compute the board's material balance, from the point of view of the "side"
// player.  This is an exact clone of the eval function in CHESS 4.5
int Board::EvalMaterial( int side )
{
  // If both sides are equal, no need to compute anything!
  if ( MaterialValue[ Player::SIDE_BLACK ] == MaterialValue[ Player::SIDE_WHITE ] )
    return 0;

  int otherSide = ( side + 1 ) % 2;
  int matTotal = MaterialValue[ side ] + MaterialValue[ otherSide ];

  // Who is leading the game, material-wise?
  if ( MaterialValue[ Player::SIDE_BLACK ] > MaterialValue[ Player::SIDE_WHITE ] )
  {
    // Black leading
    int matDiff = MaterialValue[ Player::SIDE_BLACK ] - MaterialValue[ Player::SIDE_WHITE ];
    int val = min( int(2400), matDiff ) +
                ( matDiff * ( 12000 - matTotal ) * NumPawns[ Player::SIDE_BLACK ] )
                / ( 6400 * ( NumPawns[ Player::SIDE_BLACK ] + 1 ) );
    if ( side == Player::SIDE_BLACK )
      return val;
    else
      return -val;
  } else {
    // White leading
    int matDiff = MaterialValue[ Player::SIDE_WHITE ] - MaterialValue[ Player::SIDE_BLACK ];
    int val = min( int(2400), matDiff ) +
                ( matDiff * ( 12000 - matTotal ) * NumPawns[ Player::SIDE_WHITE ] )
                / ( 6400 * ( NumPawns[ Player::SIDE_WHITE ] + 1 ) );

    if ( side == Player::SIDE_WHITE )
      return val;
    else
      return -val;
  }
}

Board::CheckStatus Board::GetCheckStatus(void)
{
  MoveListGenerator move1, move2;  
  Board Successor;  
  CheckStatus status = Normal;
  bool wayOut = false ;

  // first, compute the list of moves possible for the whole board at this time
  bool Continue = true;
  move1.ComputeLegalMoves(*this, Continue);  
  
  // then, make sure that whatever move is made does not result in checkmate  
  std::vector<Move>::iterator testMove;
  while ((testMove = move1.Next()) != move1.End())
  {    
    // make sure that moving this piece does not leave the king in check
    Successor.Clone(*this);
    Successor.ApplyMove(*testMove);
    if (move2.ComputeLegalMoves(Successor, Continue))    
      // there is a move that relieves check
      wayOut = true;
    else
      // this leaves the king in check, so it's at least check
      // unless, the move being made is by the king in which case we 
      // don't count attempting to move into check...
      if ((*testMove).MovingPiece != Board::BLACK_KING && (*testMove).MovingPiece != Board::WHITE_KING)
        status = Check;
  }
  
  // return the status
  if (wayOut)
    return status;
  else
    // none was found - checkmate
    return Checkmate;
}
