// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "ChessApplication.h"
#include "StateManager.h"
#include "InputManager.h"
#include "GameAI.h"
#include "GameOGRE.h"

// Creation/destruction
ChessApplication::ChessApplication()    
   : mStateManager(0), mGameAI(0), mInputManager(0),
     mGameOGRE(0), mShowDebugOverlay(false), mShowMessageOverlay(false), mScreenshotCounter(0) { }

ChessApplication::~ChessApplication()
{  
  delete mPlane;

  if (mGameAI!=0) 
    mGameAI->EndGame();  
  delete mInputManager;  
	delete mStateManager;        
	delete mGameAI;  
  delete mGameOGRE;
}

SceneNode* ChessApplication::GetGameRootNode()
{
  return mGameRoot;
}

// Application overrides
void ChessApplication::preStartRendering()
{
  // Setup viewports, board and lights
	createViewports();	
  createScene();
	
	// Create managers and game object
  mStateManager = new StateManager(mWindow);
  mInputManager = new ChessInputManager();
  mGameAI = new GameAI();  	    
  mGameOGRE = new GameOGRE(mPlayerCamera, mReflectCamera, mReflectCameraGround, mSceneManager, mWindow);	              
}

void ChessApplication::createViewports()
{
	// Create player camera
	mPlayerCamera = mSceneManager->createCamera("PlayerCamera");
	mPlayerCamera->setNearClipDistance(1);
	mPlayerCamera->setFarClipDistance(1000.0);    

	// Create player viewport filling the whole window
	mPlayerViewport = mWindow->addViewport(mPlayerCamera);
  mPlayerViewport->setBackgroundColour(ColourValue::Black);	
  mPlayerCamera->setAspectRatio((Real)mPlayerViewport->getActualWidth() / (Real)mPlayerViewport->getActualHeight());

  // create the camera to do the board reflection
  mReflectCamera = mSceneManager->createCamera("ReflectCam");
  mReflectCamera->setNearClipDistance(mPlayerCamera->getNearClipDistance());
  mReflectCamera->setFarClipDistance(mPlayerCamera->getFarClipDistance());
  mReflectCamera->setAspectRatio((Real)mPlayerViewport->getActualWidth() / (Real)mPlayerViewport->getActualHeight());

  // create the camera to do the ground reflection
  mReflectCameraGround = mSceneManager->createCamera("ReflectCam2");
  mReflectCameraGround->setNearClipDistance(mPlayerCamera->getNearClipDistance());
  mReflectCameraGround->setFarClipDistance(mPlayerCamera->getFarClipDistance());
  mReflectCameraGround->setAspectRatio((Real)mPlayerViewport->getActualWidth() / (Real)mPlayerViewport->getActualHeight());  	

  // Create menu viewport filling the whole window. Use the player camera to get
	// the right aspect ratio. Only displays overlays so camera isn't that important.
	mMenuViewport = mWindow->addViewport(mPlayerCamera, 2);
	mMenuViewport->setClearEveryFrame(false);
	mMenuOverlay = 0;

  // Create a message viewport in the same vein
	mMessageViewport = mWindow->addViewport(mPlayerCamera, 3);
	mMessageViewport->setClearEveryFrame(false);

  // set the overlays that will be used in-game
  OverlayManager* overlayManager = OverlayManager::getSingletonPtr();
	mMessageOverlay = static_cast<Overlay*>(overlayManager->getByName("Core/Game"));
  mDebugOverlay = static_cast<Overlay*>(overlayManager->getByName("Core/Debug"));

	// Register render target listener
	mWindow->addListener(this);
}

void ChessApplication::createScene()
{  
  // Create root scene node
  mGameRoot = mSceneManager->getRootSceneNode()->createChildSceneNode();
  mSceneManager->getRootSceneNode()->removeChild(mGameRoot);
  mRootNodeAttached = false;
  
  // load the board
  mBoard = mSceneManager->createEntity("TheBoard", "Board.mesh");      
  mBoard->getSubEntity(1)->setMaterialName("Chess/BoardSideMat");
  mGameRoot->attachObject(mBoard);                     
  
  // set up the board's reflections  
  Ogre::Plane plane(Ogre::Vector3::UNIT_Y, 0 );
  plane.d = 0;
  MeshManager::getSingleton().createPlane("board", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane, 160,160,1,1,true,1,1,1,Ogre::Vector3::UNIT_Z);

  Ogre::TexturePtr rttTexPtr = TextureManager::getSingleton().createManual("RttTex", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, TEX_TYPE_2D, 1024, 1024, MIP_DEFAULT, PF_R8G8B8, TU_RENDERTARGET);
  RenderTexture* rttTex = rttTexPtr->getBuffer()->getRenderTarget();
  Viewport *v = rttTex->addViewport(mReflectCamera);
  v->setClearEveryFrame(true);
  v->setBackgroundColour(ColourValue::Black);

  MaterialPtr mat = MaterialManager::getSingleton().create("RttMat", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
  TextureUnitState* t = mat->getTechnique(0)->getPass(0)->createTextureUnitState("Board.png");
  t = mat->getTechnique(0)->getPass(0)->createTextureUnitState("RttTex");

  // Blend with base texture
  t->setColourOperationEx(LBX_BLEND_MANUAL, LBS_TEXTURE, LBS_CURRENT, ColourValue::White, ColourValue::White, 0.25);
  t->setProjectiveTexturing(true, mReflectCamera);

  // add this object as a listener
  rttTex->addListener(this);

  // set up linked reflection
  mReflectCamera->enableReflection(plane);

  // Also clip
  mReflectCamera->enableCustomNearClipPlane(plane);

  // Give the plane a texture
  mBoard->getSubEntity(0)->setMaterialName("RttMat");              
  
  // Create a prefab plane
  mPlane = new MovablePlane("ReflectPlane");
  Entity* mPlaneEnt;
  mPlane->d = 0;
  mPlane->normal = Ogre::Vector3::UNIT_Y;
  MeshManager::getSingleton().createPlane("ReflectionPlane", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, *mPlane, 600, 600, 1, 1, true, 1, 1, 1, Ogre::Vector3::UNIT_Z);
  mPlaneEnt = mSceneManager->createEntity( "Plane", "ReflectionPlane" );

  // Attach the rtt entity to the root of the scene  
  SceneNode* mPlaneNode = mGameRoot->createChildSceneNode();

  // Attach both the plane entity, and the plane definition
  mPlaneNode->attachObject(mPlaneEnt);
  mPlaneNode->attachObject(mPlane);
  mPlaneNode->translate(0, -15, 0);  

  Ogre::TexturePtr rttTexaPtr = TextureManager::getSingleton().createManual("RttTexa", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, TEX_TYPE_2D, 1024, 1024, MIP_DEFAULT, PF_R8G8B8, TU_RENDERTARGET);
  RenderTexture* rttTexa = rttTexaPtr->getBuffer()->getRenderTarget();
  {
      Viewport *va = rttTexa->addViewport( mReflectCameraGround );
      va->setClearEveryFrame( true );
      va->setBackgroundColour( ColourValue::Black );

      MaterialPtr mata = MaterialManager::getSingleton().create("RttMata", ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
      TextureUnitState* ta = mata->getTechnique(0)->getPass(0)->createTextureUnitState("Ground.png");
      ta = mata->getTechnique(0)->getPass(0)->createTextureUnitState("RttTexa");
      mata->getTechnique(0)->getPass(0)->setSceneBlending(SBT_TRANSPARENT_ALPHA);

      // Blend with base texture
      ta->setColourOperationEx(LBX_BLEND_MANUAL, LBS_TEXTURE, LBS_CURRENT, ColourValue::White, ColourValue::White, 0.15);
      ta->setProjectiveTexturing(true, mReflectCameraGround);
      rttTexa->addListener(this);

      // set up linked reflection
      mReflectCameraGround->enableReflection(mPlane);

      // Also clip
      mReflectCameraGround->enableCustomNearClipPlane(mPlane);
  }

  // Give the plane a texture
  mPlaneEnt->setMaterialName("RttMata");  

  // set the ambient light
  mSceneManager->setAmbientLight(ColourValue(0.9, 0.6, 0.2));

  // and the other lights
  Light* l;
  l = mSceneManager->createLight("Light1");
  l->setPosition(-200, 200, -200);
  l->setDiffuseColour(0.5, 0.5, 0.5);
  l->setSpecularColour(0.7, 0.7, 0.7);
  l->setCastShadows(false);

  l = mSceneManager->createLight("Light2");
  l->setPosition(-200, 200, 200);
  l->setDiffuseColour(0.5, 0.5, 0.5);
  l->setSpecularColour(0.7, 0.7, 0.7);
  l->setCastShadows(false);

  l = mSceneManager->createLight("Light3");
  l->setPosition(200, 200, -200);
  l->setDiffuseColour(0.5, 0.5, 0.5);
  l->setSpecularColour(0.7, 0.7, 0.7);
  l->setCastShadows(false);

  l = mSceneManager->createLight("Light4");
  l->setPosition(200, 200, 200);
  l->setDiffuseColour(0.5, 0.5, 0.5);
  l->setSpecularColour(0.7, 0.7, 0.7);
  l->setCastShadows(true);    
}

void ChessApplication::setMenuOverlay(Overlay* menuOverlay)
{
	mMenuOverlay = menuOverlay;
}

bool ChessApplication::getShowDebugOverlay()
{
	return mShowDebugOverlay;
}

void ChessApplication::setShowDebugOverlay(bool show)
{
	mShowDebugOverlay = show;
}

void ChessApplication::setShowMessageOverlay(bool show)
{
  mShowMessageOverlay = show;
}

void ChessApplication::updateDebugOverlay()
{
	if (mShowDebugOverlay)
	{
		const RenderTarget::FrameStats& stats = mWindow->getStatistics();
    OverlayManager* overlayManager = OverlayManager::getSingletonPtr();

		// Updating on changes only improves FPS dramatically!
		if (stats.lastFPS != mLastFrameStats.lastFPS)
			overlayManager->getOverlayElement("Core/CurrentFPS")->setCaption("Current FPS: " + StringConverter::toString(stats.lastFPS));

		if (stats.triangleCount != mLastFrameStats.triangleCount)
			overlayManager->getOverlayElement("Core/Triangles")->setCaption("Triangles: " + StringConverter::toString(stats.triangleCount));

		mLastFrameStats = stats;
	}
}

void ChessApplication::saveScreenshot()
{
	mWindow->writeContentsToTimestampedFile("chess", ".png");  
}

// RenderTargetListener overrides
void ChessApplication::preViewportUpdate(const RenderTargetViewportEvent &evt)
{ 
	if (evt.source == mPlayerViewport)
	{   
    // Create a skybox
    mSceneManager->setSkyBox(true, "Chess/Sunset", 50);		

    AttachRootNode();
	}	
  else if (evt.source == mMessageViewport)
  {
    // Show debug overlay if needed
		if (mShowDebugOverlay)
			mDebugOverlay->show();  

    // Show game message if needed
    if (mShowMessageOverlay)
      mMessageOverlay->show();
  }
	else if (evt.source == mMenuViewport)
	{
		// Setup for rendering menu viewport
		if (mMenuOverlay)
			mMenuOverlay->show();
	}
}

void ChessApplication::postViewportUpdate(const RenderTargetViewportEvent &evt)
{  
	if (evt.source == mPlayerViewport)
	{		
    // Turn off a da skybox
    mSceneManager->setSkyBox(false, "Chess/Sunset", 50);	   
    DetachRootNode();
	}  
  else if (evt.source == mMessageViewport)
  {
    // Hide debug overlay    
		if (mShowDebugOverlay)
			mDebugOverlay->hide();	

    // Hide game message if needed
    if (mShowMessageOverlay)
      mMessageOverlay->hide();
  }
	else if (evt.source == mMenuViewport)
	{
		// Cleanup after rendering menu viewport
		if (mMenuOverlay)
			mMenuOverlay->hide();
	}  
}

// render target events
void ChessApplication::preRenderTargetUpdate(const RenderTargetEvent& evt)
{
  mStateManager->capture();
  AttachRootNode();
}

void ChessApplication::postRenderTargetUpdate(const RenderTargetEvent& evt)
{    
  DetachRootNode();
}

void ChessApplication::AttachRootNode()
{
  // Cleanup after rendering player viewport
  if (!mRootNodeAttached)
  {
    mSceneManager->getRootSceneNode()->addChild(mGameRoot);
    mRootNodeAttached = true;
  }
}

void ChessApplication::DetachRootNode()
{
  // Cleanup after rendering player viewport
  if (mRootNodeAttached)
  {
    mSceneManager->getRootSceneNode()->removeChild(mGameRoot);
    mRootNodeAttached = false;
  }
}
