// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "GameAI.h"
#include "Board.h"
#include "Piece.h"
#include "Hand.h"
#include "Move.h"
#include "Player.h"
#include "PlayerAI.h"
#include "PlayerHuman.h"
#include "GameOGRE.h"
#include "AISearchAgent.h"
#include "StateManager.h"
#include "HandState.h"
#include "PromotionMenuState.h"
#include "GameOverState.h"
#include "ChessApplication.h"

// Singleton instance
template<> GameAI* Singleton<GameAI>::ms_Singleton = 0;

Board* GameAI::GetGameBoard(void) {return GameBoard;}

GameAI::GameAI() : IsThreadRunning(false), IsRunning(false), 
                   nextMove(0), mDifficulty("Easy"), AIThread(0)
{
  // initialise these objects      
  Players[0] = 0;
  Players[1] = 0;
  GameBoard = new Board;   
}

GameAI::~GameAI() 
{
  Piece::DestroyPieces();
  Hand::DestroyHands();
  EndGame();
  // clean up  
  delete nextMove;
  delete AIThread;
  delete Players[0];
  delete Players[1];
  delete GameBoard;  
}

bool GameAI::InitializeGame() 
{    
  delete nextMove;
  nextMove = 0;
  GameBoard->StartingBoard();   
  Piece::SetupPieces(GameBoard);    
  Hand::CreateHands();
  
  return true;    
}

void GameAI::SetPlayer(const int colour, const int type)
{
  // clear out the old one
  delete Players[colour]; 

  // create a new one
  if (type == 0)
  {
    Players[colour] = new PlayerAI(colour, AISearchAgent::AISEARCH_MTDF);
    static_cast<PlayerAI*>(Players[colour])->SetDifficulty(mDifficulty);
  } 
  else
    Players[colour] = new PlayerHuman(colour);
}

void GameAI::Start(void)
{
  IsRunning = true;  
  GameOGRE::getSingletonPtr()->Pause(false);
}

void GameAI::Pause(void)
{
  IsRunning = false;  
  GameOGRE::getSingletonPtr()->Pause(true);
}

void GameAI::EndGame(void)
{  
  IsRunning = false; 
  GameOGRE::getSingletonPtr()->Pause(true);
  DestroyAIThread();
}

void GameAI::Update()
{     
  if (!IsRunning) return;  

  // see if we have a move to make
  if (nextMove!=0)
  {
    // commission the move to the game board
    GameBoard->ApplyMove(*nextMove, true);            

    // clear out the last move
    DestroyAIThread();
  }

  // only start a move when both hands have finished moving
  if (Hand::GetHand(Player::SIDE_WHITE)->GetHandState()->GetStateType() == HandState::Static
      && Hand::GetHand(Player::SIDE_BLACK)->GetHandState()->GetStateType() == HandState::Static            
      && Players[GameBoard->GetCurrentPlayer()]->GetType() == Player::TYPE_AI
      && !IsThreadRunning)
      {    
        // create a new AI thread    
        CreateAIThread();
      }   
}

void GameAI::CreateAIThread(void)
{  
  // create the thread  
  IsThreadRunning = true;
  AIThread = new boost::thread(boost::bind(&GameAI::AITurn, this));
}

void GameAI::DestroyAIThread(void)
{  
  // clean up the last thread
  if (AIThread!=0)
  {
    IsThreadRunning = false; // allows us to leave the AI loop quicker
    AIThread->join();
    delete AIThread;
    AIThread = 0;    
  } 
  delete nextMove;
  nextMove = 0;  
}

void GameAI::AITurn(void)
{      
  // Ask the next player for a move
  nextMove = new Move(Players[GameBoard->GetCurrentPlayer()]->GetMove(*GameBoard, IsThreadRunning));  
}

bool GameAI::IsCurrentPlayerAI(void)
{
  // let's us know if the current player is an AI player.
  return (GetCurrentPlayer()->GetType() == Player::TYPE_AI);
}

int GameAI::GetCurrentPlayerColour(void)
{
  return GameBoard->GetCurrentPlayer();
}

Player* GameAI::GetCurrentPlayer(void)
{
  return Players[GameBoard->GetCurrentPlayer()];
}

void GameAI::CreateMove(int source, int dest)
{        
  testMove = new Move();
  testMove->SourceSquare = source;
  testMove->DestinationSquare = dest;  

  // only set the move if we've not got to decide on a promotion
  if (static_cast<PlayerHuman*>(Players[GameBoard->GetCurrentPlayer()])->CompileMove(GameBoard, testMove))
    nextMove = testMove;
  else
    // show the promotion menu
    StateManager::getSingleton().pushState(PromotionMenuState::getSingletonPtr());
}

void GameAI::SetPromotion(String type)
{  
  // set the promotion
  if (type == "Queen")
    testMove->MoveType += Move::MOVE_PROMOTION_QUEEN;
  else if (type == "Rook")
    testMove->MoveType += Move::MOVE_PROMOTION_ROOK;
  else if (type == "Bishop")
    testMove->MoveType += Move::MOVE_PROMOTION_BISHOP;
  else
    testMove->MoveType += Move::MOVE_PROMOTION_KNIGHT;    

  // finish setting the next move
  nextMove = testMove;  
}

void GameAI::GetCheckStatus(void)
{
  // see if it's checkmate
  Board::CheckStatus status = GameBoard->GetCheckStatus();

  switch (status)
  {
    case Board::Normal:
      // hide the check message
      static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->setShowMessageOverlay(false);
      break;
    case Board::Check:
      // display the check message
      static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->setShowMessageOverlay(true);
      break;
    case Board::Checkmate:
      // display the winner screen
      if (GameBoard->GetCurrentPlayer() == Player::SIDE_WHITE) {
        GameOverState::getSingletonPtr()->DisplayResult("Checkmate!", "Black Wins!");            
      } else {
        GameOverState::getSingletonPtr()->DisplayResult("Checkmate!", "White Wins!");            
      }
      StateManager::getSingleton().pushState(GameOverState::getSingletonPtr());
    break;
  }
}

void GameAI::SetDifficulty(String difficulty)
{
  mDifficulty = difficulty;
}