// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "GameState.h"
#include "CollisionManager.h"
#include "ChessApplication.h"
#include "InputManager.h"
#include "GameAI.h"
#include "GameOGRE.h"
#include "GameMenuState.h"
#include "Piece.h"
#include "StateManager.h"

GameState::GameState() : mLMouseDown(false), mRMouseDown(false), mPieceSelected(false)
{
  // get some pointers  
  mSceneManager = static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->getSceneManager();    
  mInputManager = ChessInputManager::getSingletonPtr();

  // Create RaySceneQuery
  mRaySceneQuery = mSceneManager->createRayQuery(Ogre::Ray());
  mRaySceneQuery->setSortByDistance(true);

  // Create a plane
  mPlane = new Ogre::Plane(Ogre::Vector3::UNIT_Y, 0);
}

GameState::~GameState()
{
  delete mPlane;
  delete mRaySceneQuery;
}

// State overrides
void GameState::pause()
{  
  // stop the AI thread if it is running
  mGameAI->Pause();  

	// Save game camera yaw and pitch so it can be restored, when game continues
	mSavedCameraYaw = mGameOGRE->mCameraWantedYaw;
	mSavedCameraPitch = mGameOGRE->mCameraWantedPitch;
}

void GameState::resume()
{  
  // restart the AI thread
  mGameAI->Start();  

	// Restore saved game camera yaw and pitch
	mGameOGRE->mCameraWantedYaw = mSavedCameraYaw;
	mGameOGRE->mCameraWantedPitch = mSavedCameraPitch;
}

void GameState::keyPressed(const KeyEvent & e)
{
	// Handle keys common to all game states
	switch (e.key)
	{
	case KC_ESCAPE:    
		StateManager::getSingleton().pushState(GameMenuState::getSingletonPtr());
		break;

	default:
		// Default processing
		ChessState::keyPressed(e);
	}
}

void GameState::mouseMoved(const MouseEvent & e)
{
  // update the mouse pointer
  ChessState::mouseMoved(e);

  static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->AttachRootNode();

  // Adjust and clamp wanted camera distance
  const Real minDistance = 22.5;
  const Real maxDistance = 500;

	mGameOGRE->mCameraWantedDistance -= e.state.Z.rel;

	if (mGameOGRE->mCameraWantedDistance < minDistance)
		mGameOGRE->mCameraWantedDistance = minDistance;

  if (mGameOGRE->mCameraWantedDistance > maxDistance)
		mGameOGRE->mCameraWantedDistance = maxDistance;  

  if (mRMouseDown)
  {  
    const Real maxYaw = 0.45*Math::PI;
    const Real maxPitch = 0.05*Math::PI;
    const Real minPitch = 0.45*Math::PI;	  

    // Adjust and clamp wanted camera yaw
    mGameOGRE->mCameraWantedYaw -= e.state.X.rel;
    
    // Adjust and clamp wanted camera pitch
    mGameOGRE->mCameraWantedPitch -= e.state.Y.rel;
    
    if (mGameOGRE->mCameraWantedPitch > -Radian(maxPitch).valueDegrees()) {
      mGameOGRE->mCameraWantedPitch = -Radian(maxPitch).valueDegrees();
    } else if (mGameOGRE->mCameraWantedPitch < -Radian(minPitch).valueDegrees()) {
      mGameOGRE->mCameraWantedPitch = -Radian(minPitch).valueDegrees();
    }	    
  }    

  // move the current position
  if (!mGameAI->IsCurrentPlayerAI() && mPieceSelected)
  {
    // create a ray    
    Ogre::Ray mouseRay = mGameOGRE->CreateRay();
          
    // get the intersection with the plane
    std::pair<bool, Real> point = mouseRay.intersects(*mPlane);        
    if (point.first)
      mInputManager->ShowCurrentNode(mouseRay.getPoint(point.second), mSelectedPiece, mCurrentPosition);
  }

  // hide / show the mouse pointer  
  if (mGameAI->IsCurrentPlayerAI() || mRMouseDown)    
    CEGUI::MouseCursor::getSingleton().hide();
  else
    CEGUI::MouseCursor::getSingleton().show();    

  static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->DetachRootNode();
}

void GameState::mousePressed(const MouseEvent & e)
{
  if (e.state.buttonDown(MB_Left))
    mLMouseDown = true;
  else
    mLMouseDown = false;

  if (e.state.buttonDown(MB_Right))
    mRMouseDown = true;
  else
    mRMouseDown = false;    

  static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->AttachRootNode();

  // only show the selection fella if it's the players turn and they're not readjusting the view
  if (!mGameAI->IsCurrentPlayerAI() && mLMouseDown)
  {    
    // create a ray    
    Ogre::Ray mouseRay = mGameOGRE->CreateRay(); 

    // see if a move is being made
    if (mPieceSelected)
    {                  
      // hide the current position node
      mInputManager->HideCurrentNode();
      // get the intersection with the plane
      std::pair<bool, Real> point = mouseRay.intersects(*mPlane);        
      if (point.first)
      {
        if (mSelectedPiece->IsMoveValid(mCurrentPosition))
        {
          // feed the move into the game AI
          GameAI::getSingletonPtr()->CreateMove(mSelectedPiece->GetPosition(), mCurrentPosition);
          // hide the rest
          mInputManager->HideSelectionNode();          
          mPieceSelected = false;
          mSelectedPiece = 0;
          return;
        }
      }
    }    

    // hide the selection node to start
    mInputManager->HideSelectionNode();
    mPieceSelected = false;
    mSelectedPiece = 0;

    // create a rayscenequery    
    mRaySceneQuery->setRay(mouseRay);       

    // see if we're hovering over a piece
    RaySceneQueryResult &result = mRaySceneQuery->execute();
    RaySceneQueryResult::iterator itr;
              
    for (itr = result.begin(); itr != result.end(); itr++)     
    {      
      if (itr->movable->getName().substr(0,5) == "Piece") 
      {        
        // drill a bit deeper and check for an accurate collision
        Piece* thePiece = Piece::GetPiece(itr->movable->getName());
        if (thePiece->GetColour() == mGameAI->GetCurrentPlayerColour())
        {                    
          if(CollisionManager::getSingleton().collide(&mouseRay, thePiece->GetCollisionEntity()))
          {
            // record this piece
            mInputManager->ShowSelectionNode(thePiece->GetSceneNode());
            mSelectedPiece = thePiece;  
            thePiece->GetValidMoves();
            mPieceSelected = true;
            break;
          } 
        }         
      }      
    }   
  }  
  static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->DetachRootNode();
}

void GameState::mouseReleased(const MouseEvent & e)
{
  if (!e.state.buttonDown(MB_Left))
    mLMouseDown = false;  

  if (!e.state.buttonDown(MB_Right))
    mRMouseDown = false;   
}

void GameState::frameStarted(Real dt)
{
  // update the input manager
  mInputManager->Update(dt);

  // do base class
  ChessState::frameStarted(dt);
}
