// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "Hand.h"
#include "Application.h"
#include "StaticState.h"
#include "PickupState.h"
#include "MovePieceState.h"
#include "CaptureState.h"
#include "RemoveState.h"
#include "ReturnState.h"
#include "PromotionState.h"
#include "ChessApplication.h"

Hand* Hand::mHands[2];
bool Hand::IsInitialised = false;

SceneManager* Hand::GetSceneManagerPtr(void) {return mSceneManager;}
Entity* Hand::GetEntityPtr(void) {return mEntity;}
SceneNode* Hand::GetSceneNodePtr(void) {return mScenenode;}  
HandState* Hand::GetHandState(void) {return mHandState;}

Hand::Hand(const int colour) : mColour(0), mEntity(0), mScenenode(0), mHandState(0), PiecesTaken(0), mSceneManager(0)
{
  // keep a pointer to the scene manager for quickness-sake
  mSceneManager = static_cast<Application*>(Application::getSingletonPtr())->getSceneManager();      

  // set which colour this hand represents
  mColour = colour;  

  // create an entity  
  mEntity = mSceneManager->createEntity("Hand" + StringConverter::toString(mColour), "Hand.mesh");        

  // attach it to the scene node  
  mScenenode = static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->GetGameRootNode()->createChildSceneNode("Hand" + StringConverter::toString(mColour));
  mScenenode->attachObject(mEntity);              

  // if it's the black hand, rotate it 180
  if (mColour) {        
    mScenenode->setOrientation(Quaternion::IDENTITY);
    mScenenode->yaw(Degree(180));    
  }   

  // set the material and the shadow casting ...
  // ...we're not using shadows yet but in case we change our minds...
  mEntity->setMaterialName("Chess/Hand");  
  mEntity->setCastShadows(false);       

  // set up the particle effects    
  mMagicWand = Application::getSingletonPtr()->getSceneManager()->createParticleSystem("MagicWand" + StringConverter::toString(mColour), "MagicWand");      
  mScenenode->attachObject(mMagicWand);        

  // set the hand into a waiting state  
  SetAction(HandState::Waiting);
}

Hand::~Hand()
{
  // get rid of the hand state  
  delete mHandState;

  // Destroy scene node
  if (mScenenode) {
	  mScenenode->detachAllObjects();
	  mScenenode->getParentSceneNode()->removeAndDestroyChild(mScenenode->getName());
    mScenenode = 0;
  }

  // destroy the particle system
  if (mMagicWand) {
    Application::getSingletonPtr()->getSceneManager()->destroyParticleSystem(mMagicWand);
    mMagicWand = 0;
  }

	// Destroy entity
	if (mEntity) {
		mSceneManager->destroyEntity(mEntity);    
    mEntity = 0;
  }
}

// this is called when a new game starts - puts the hands back to original settings
void Hand::CreateHands(void)
{  
  // get rid of previous hands
  if (IsInitialised)
    DestroyHands();

  // create new hands
  for (int i = 0; i < 2; i++) {        
    mHands[i] = new Hand(i);  
  }  

  // reset these values  
  IsInitialised = true;
}

// can be called when the program closes down and when a new game begins
void Hand::DestroyHands(void)
{  
  // delete any previous hands    
  for (int i = 0; i < 2; i++)  
    delete mHands[i];  
}

// this is called when the hands need to perform an action
void Hand::SetAction(HandState::ActionType action, 
                     const int Source, const int Destination,
                     const int CastlingSource, const int CastlingDestination,
                     const String PromoType)
{  
  // create a temporary state - this could be more elegant using casting, but I couldn't get it to work - 
  // I think I know how to do it now and will pursue it further next time
  delete mHandState;
  mHandState = new HandState(this, action, Source, Destination, CastlingSource, CastlingDestination, PromoType);    

  // set the starting state for this action
  switch (action)
  {
    case HandState::Waiting:      
      SetState(HandState::Static);      
      break;
    case HandState::Move:            
    case HandState::Capture:      
    case HandState::Castle:
    case HandState::EnPassant:      
    case HandState::Promotion:    
      SetState(HandState::Pickup);      
      break;        
  }    
}

// this is a subset state of the current action
void Hand::SetState(const HandState::StateType state)
{  
  // call the initialise routine for this state
  switch (state)
  {    
    case HandState::Static:
      static_cast<StaticState*>(mHandState)->Initialise();      
      break;
    case HandState::Pickup:      
      static_cast<PickupState*>(mHandState)->Initialise();
      break;
    case HandState::MovePiece:      
      static_cast<MovePieceState*>(mHandState)->Initialise();          
      break;
    case HandState::Take:
      static_cast<CaptureState*>(mHandState)->Initialise();
      break;
    case HandState::Remove:
      static_cast<RemoveState*>(mHandState)->Initialise();
      break;
    case HandState::SimpleReturn:
    case HandState::CaptureReturn:
      static_cast<ReturnState*>(mHandState)->Initialise();
      break;  
    case HandState::NewPiece:
      static_cast<PromotionState*>(mHandState)->Initialise();
      break;
  }           
}

// turn the particles on and off
void Hand::ToggleParticle(bool toggle)
{
  for (int i = 0; i < mMagicWand->getNumEmitters(); i++)
    mMagicWand->getEmitter(i)->setEnabled(toggle);
}

// called once each frame by the main loop 
// - this moves the hands if necessary and consequently, the pieces
void Hand::UpdateHands(const Real dt)
{
  if (!IsInitialised) return;
  // cycle through the hands
  for (int i = 0; i < 2; i++)    
  {
    // find out which state we are in - reflection(ish)
    HandState::StateType currentState = mHands[i]->mHandState->GetStateType();
    HandState::StateType nextState;

    // call the relevant update routine
    switch (currentState)
    {    
      case HandState::Static:
        nextState = static_cast<StaticState*>(mHands[i]->mHandState)->Update(dt);      
        break;
      case HandState::Pickup:      
        nextState = static_cast<PickupState*>(mHands[i]->mHandState)->Update(dt);
        break;
      case HandState::MovePiece:      
        nextState = static_cast<MovePieceState*>(mHands[i]->mHandState)->Update(dt);          
        break;
      case HandState::Take:
        nextState = static_cast<CaptureState*>(mHands[i]->mHandState)->Update(dt);
        break;
      case HandState::Remove:
        nextState = static_cast<RemoveState*>(mHands[i]->mHandState)->Update(dt);
        break;
      case HandState::SimpleReturn:
      case HandState::CaptureReturn:
        nextState = static_cast<ReturnState*>(mHands[i]->mHandState)->Update(dt);
        break;  
      case HandState::NewPiece:
        nextState = static_cast<PromotionState*>(mHands[i]->mHandState)->Update(dt);
        break;
    }               
    // change states if the last one has ended
    if (currentState != nextState)
      mHands[i]->SetState(nextState);        
  }
}
