// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "PickupState.h"
#include "Hand.h"
#include "Piece.h"
#include "AnimationBlender.h"

void PickupState::Initialise(void) 
{ 
  // record this state type
  mStateType = HandState::Pickup;   

  // turn on particles
  mHand->ToggleParticle(true);

  // it will take one second from static to touching piece
  mStateDuration = 1; 
  
  // if we're castling, put the first piece back down
  // and record the next piece
  if ((mActionType == HandState::Castle && mCastleFirst) ||
       mActionType == HandState::EnPassant && mEnPassantFirst)
  {
    // update the piece structure to its new position
    mMovingPiece->SetPosition(mDestinationSquare);
    // if we're castling, now grab the rook
    if (mActionType == HandState::Castle)
    {             
      mSourceSquare = mCastlingSource;
      mDestinationSquare = mCastlingDestination;  
    }
    // then detach and get the second piece
    DetachPiece(mMovingPiece, mMPSceneNode);
    mMovingPiece = Piece::GetPiece(mSourceSquare);    
  }  

  // put the hand in it's starting position
  CreateMovementTrack();  

  // set the initial animation
  StartAnimation();  
}

void PickupState::CreateMovementTrack()
{    
  // get the world coordinates of the piece being picked up
  mTargetPosition = Ogre::Vector3::ZERO;
  Functions::ConvertCoords(mSourceSquare, mTargetPosition.x, mTargetPosition.z);  

  // tweak the offsets
  mOffset.x = 32;
  mOffset.y = 50;
  mOffset.z = 62.5;

  // make sure the scene node's rotation is taken into account
  Ogre::Matrix3 nodeRot;      
  mHandNode->_getDerivedOrientation().ToRotationMatrix(nodeRot);  
  mOffset = mOffset * nodeRot;

  // tweak the position according to the offsets
  mTargetPosition += mOffset;  

  // store the movement vector (in world coordinates) 
  // of the target position from the entities current position
  mMovementVector = mTargetPosition - mHandNode->getPosition();  
      
  // create the path to be followed     
  Animation* anim = mSceneManager->createAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()), mStateDuration);

  // Spline it for nice curves
  anim->setInterpolationMode(Animation::IM_SPLINE);

  // Create a track to animate the hand
  NodeAnimationTrack* track = anim->createNodeTrack(0, mHandNode);

  // Setup keyframes
  TransformKeyFrame* key = track->createNodeKeyFrame(0); // startposition
  key->setTranslate(mHandNode->getPosition());
  key->setRotation(mHandNode->getOrientation());  
  
  key = track->createNodeKeyFrame(mStateDuration * 0.25);
  key->setTranslate(Ogre::Vector3(mTargetPosition.x - (mMovementVector.x * 0.85), mTargetPosition.y + 25, mTargetPosition.z - (mMovementVector.z * 0.85)));
  key->setRotation(mHandNode->getOrientation());  

  key = track->createNodeKeyFrame(mStateDuration * 0.75);
  key->setTranslate(Ogre::Vector3(mTargetPosition.x - (mMovementVector.x * 0.15), mTargetPosition.y + 35, mTargetPosition.z - (mMovementVector.z * 0.15)));
  key->setRotation(mHandNode->getOrientation());    

  key = track->createNodeKeyFrame(mStateDuration);
  key->setTranslate(mTargetPosition);  
  key->setRotation(mHandNode->getOrientation()); 
}

void PickupState::StartAnimation(void)
{
  // start the movement track
  HandState::BaseStartAnimation();

  // run the pickup animation
  mAnimationBlender->blend("Pickup",  AnimationBlender::BlendWhileAnimating, mStateDuration * 0.1, false);   
}

HandState::StateType PickupState::Update(Real dt)
{         
  // cycle the animation
  HandState::BaseUpdate(dt);  

  // if we reach the end of this state's length, move onto the next one.
  if (mMoveState->getTimePosition() >= mStateDuration)
  {
    // make sure that the hand has completely finished moving
    mHandNode->setPosition(mTargetPosition);

    mSceneManager->destroyAnimation("HandTrack" + StringConverter::toString(mHand->GetColour()));
    mMoveState = 0;

    if (mActionType == HandState::Move ||
        mActionType == HandState::Castle ||
        mActionType == HandState::EnPassant ||
        mActionType == HandState::Promotion)      
      return HandState::MovePiece;  
    else
     return HandState::Take;
  }
  else    
    return mStateType;
}
