// Chess 
// Copyright (C) 2005  Steve Driscoll (steviedisco@gmail.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA

#include "stdafx.h"
#include "Piece.h"
#include "ChessApplication.h"
#include "CollisionEntity.h"
#include "CollisionManager.h"
#include "Board.h"
#include "Player.h"
#include "PlayerHuman.h"
#include "GameAI.h"

// set up static variables
int Piece::mPieceNumber = 0;
Piece* Piece::mPieces[32];
SceneManager* Piece::mSceneManager = 0;

CollisionEntity* Piece::GetCollisionEntity(void) {return mColEntity;}

Piece::Piece(const int colour, const String type, const int position, const int rotation, const String shortcode) 
    : mOrigin(Ogre::Vector3::ZERO), mDestination(Ogre::Vector3::ZERO)// , mColEntity(0)
{
  // get a pointer for quickness
  mSceneManager = static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->getSceneManager();  
  
  // set the piece's position and rotation  
  mPosition = position;  
  mShortCode = shortcode;

  // create a scene node, position it
  Functions::ConvertCoords(mPosition, (mOrigin.x = mDestination.x), (mOrigin.z = mDestination.z));  
  mScenenode = static_cast<ChessApplication*>(ChessApplication::getSingletonPtr())->GetGameRootNode()->createChildSceneNode("Piece" + StringConverter::toString(mPieceNumber), mOrigin);           

  // load a mesh and attach the entity  
  mEntity = mSceneManager->createEntity("Piece" + StringConverter::toString(mPieceNumber), type + ".mesh");        
  mScenenode->attachObject(mEntity);           

  // set the material
  mColour = colour;
  if (mColour == Player::SIDE_WHITE) 
    mEntity->setMaterialName("Chess/White");
  else
    mEntity->setMaterialName("Chess/Black");
  mEntity->setCastShadows(true);   

  // create an entity to check for collisions  
  mColEntity = CollisionManager::getSingleton().createEntity(mEntity);  

  // rotate the piece if necessary
  if (rotation!=0) {
    mScenenode->setOrientation(Quaternion::IDENTITY);
    mScenenode->yaw(Degree(rotation));
  }

  // increment the static count to the next piece
  mPieceNumber++;
}

// destructor
Piece::~Piece()
{
  // get rid of the collision entity  
  delete mColEntity;

  // Destroy scene node
  if (mScenenode)
  {
	  mScenenode->detachAllObjects();
	  mScenenode->getParentSceneNode()->removeAndDestroyChild(mScenenode->getName());
  }

	// Destroy entity
	if (mEntity)
		mSceneManager->destroyEntity(mEntity);
}

// set up the pieces according to the board being passed in
void Piece::SetupPieces(Board* theBoard)
{ 
  // clear out the old array
  if (mPieceNumber != 0)
    DestroyPieces();

  mPieceNumber = 0;
    
  for( int line = 0; line < 8; line++ )
  {    
    for( int col = 0; col < 8; col++ )
    {
      dong bits = Board::SquareBits[ line * 8 + col ];

      // Scan the bitboards to find a piece, if any
      int piece = 0;
      while ((piece < Board::ALL_PIECES) && ((bits & theBoard->BitBoards[piece]) == 0))
        piece++;

      // One exception: don't show the "phantom kings" which the program places
      // on the board to detect illegal attempts at castling over an attacked
      // square
      if ( ( piece == Board::BLACK_KING ) && ( ( theBoard->ExtraKings[ Player::SIDE_BLACK ] & Board::SquareBits[ line * 8 + col ] ) != 0 ) )
        piece = Board::EMPTY_SQUARE;
      if ( ( piece == Board::WHITE_KING ) && ( ( theBoard->ExtraKings[ Player::SIDE_WHITE ] & Board::SquareBits[ line * 8 + col ] ) != 0 ) )
        piece = Board::EMPTY_SQUARE;

      // This gets incremented by the constructor, compilers seem to differ in
      // opinion wether the post or pre incremeted value should be used in
      // the switch case below, this works around this.
      int old_mPieceNumber=mPieceNumber;
      // Create the piece
      switch (piece)
      {
        case Board::BLACK_KING:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "King", (line * 8) + col, 0, "BK");
          break;
        case Board::BLACK_QUEEN:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Queen", (line * 8) + col, 0, "BQ");
          break;
        case Board::BLACK_BISHOP:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Bishop", (line * 8) + col, 90, "BB");
          break;
        case Board::BLACK_KNIGHT:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Knight", (line * 8) + col, 180, "BN");
          break;        
        case Board::BLACK_ROOK:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Rook", (line * 8) + col, 0, "BR");
          break;
        case Board::BLACK_PAWN:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Pawn", (line * 8) + col, 0, "BP");
          break;
        case Board::WHITE_KING:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "King", (line * 8) + col, 0, "WK");
          break;
        case Board::WHITE_QUEEN:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Queen", (line * 8) + col, 0, "WQ");
          break;
        case Board::WHITE_BISHOP:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Bishop", (line * 8) + col, 90, "WB");
          break;
        case Board::WHITE_KNIGHT:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Knight", (line * 8) + col, 0, "WN");
          break;
        case Board::WHITE_ROOK:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Rook", (line * 8) + col, 0, "WR");
          break;
        case Board::WHITE_PAWN:
          mPieces[old_mPieceNumber] = new Piece((piece % 2), "Pawn", (line * 8) + col, 0, "WP");
          break;
      }      
    }    
  }  
}


// get rid of the static array of pieces
void Piece::DestroyPieces(void)
{
  for (int i = 0; i < 32; i++)
    delete mPieces[i];
}

// promote this piece - change it's entity basically
void Piece::PromotePiece(const String type)
{
  // record the name
  String name = mEntity->getName();

  // detach and remove the old mesh
  mScenenode->detachObject(mEntity);
  mSceneManager->destroyEntity(mEntity);

  // load and attach a new mesh  
  mEntity = mSceneManager->createEntity(name, type + ".mesh");
  if (mColour == Player::SIDE_WHITE) 
    mEntity->setMaterialName("Chess/White");
  else
    mEntity->setMaterialName("Chess/Black");
  mScenenode->attachObject(mEntity);  

  // get rid of the previous collision entity  
  delete mColEntity;

  // create an entity to check for collisions
  mColEntity = CollisionManager::getSingleton().createEntity(mEntity);  
}

// find a piece by its position on the board
Piece* Piece::GetPiece(const int position)
{
  for (int i = 0; i < 32; i++) {
    if (mPieces[i]->mPosition == position)
    {      
      return mPieces[i];
    }
  }    
  return 0; // no piece
}

// find a piece by its name
Piece* Piece::GetPiece(const String name)
{
  for (int i = 0; i < 32; i++) {
    if (mPieces[i]->GetEntity()->getName() == name)
    {      
      return mPieces[i];
    }
  }    
  return 0; // no piece
}

// sets the position of this piece on the board (square 00 - 63)
void Piece::SetPosition(const int position)
{  
  mPosition = position;
  Functions::ConvertCoords(mPosition, mDestination.x, mDestination.z);    
  mOrigin = mDestination - mScenenode->getPosition();      
}

// gets a list of valid moves for this piece
void Piece::GetValidMoves(void)
{
  GameAI* game = GameAI::getSingletonPtr();  
  mValidMoves = static_cast<PlayerHuman*>(game->GetCurrentPlayer())->GetValidMoves(*(game->GetGameBoard()), mPosition);  
}

// test whether this move is valid for this piece
bool Piece::IsMoveValid(int destPosition)
{
  std::vector<int>::iterator testPos;
  for (testPos = mValidMoves.begin(); testPos != mValidMoves.end(); testPos++)
  {
    if (*testPos == destPosition)
      return true;
  }
  return false;
}
